# coding=utf-8
from __future__ import print_function, unicode_literals

import logging
import re
from datetime import datetime

import boto3

JOB_Q = ''  # etc etc

CLOUDWATCH_NS = 'AtomQueueWorker'
DEFAULT_CONTAINER_IMAGE = 'ubuntu:18.04'
DEFAULT_JOB_SIZE = 'small'

JOB_SIZE_ATTRIB = 'Atom.JobSize'
"""The ``MessageAttribute`` name in which one will find a key from :ref:`TASKS_BY_JOB_SIZE`"""

TASK_OVERRIDE_ATTRIB = 'Atom.TaskArn'
"""The ``MessageAttribute`` name in which one may override the Task ARN"""

SCRIPTING_CONTAINER_NAME = 'scripting'
"""The name of the container in the Task to which we will submit scripts"""

# fill this in, perhaps even dynamically
ECS_CLUSTER_ID = ''

TASKS_BY_JOB_SIZE = {
    'small': 'urn:alpha-beta:task1234',
    'medium': 'urn:alpha-beta:task4567',
}

_cloudwatch_metric_template = {
    'Namespace': CLOUDWATCH_NS,
    'MetricData': [
    ]
}

ecs = boto3.client('ecs')
sqs = boto3.client('sqs')
cloud_watch = boto3.client('cloudwatch')


def publish_cloudwatch(metric_name, value, **dimensions):
    """
    :param str metric_name:
    :param float value:
    :param str dimensions: any metric dimensions to associate
    """
    # -----^^^ evidently for kwargs, the `dict[str,` is implied,
    # and we're only documenting the value-type
    cwm = dict(_cloudwatch_metric_template)
    m = {
        'MetricName': metric_name,
        'Timestamp': datetime.utcnow(),
        'Value': value,
    }
    if dimensions:
        m['Dimensions'] = [{'Name': k, 'Value': v}
                           for k, v in dimensions.items()]
    cwm['MetricData'].append(m)
    cloud_watch.put_metric_data(**cwm)


def on_message(m):
    m_id = m['MessageId']
    # *attempt* to associate the RunTask with its MID
    task_link = re.search(
        '.{1,36}', re.sub(r'[^A-Za-z0-9_-]', '', m_id)).group(0)
    r_handle = m['ReceiptHandle']

    #: :type: dict[str,dict]
    ma = m['MessageAttributes']
    job_size = ma.get(JOB_SIZE_ATTRIB, DEFAULT_JOB_SIZE)
    task_arn = ma.get(TASK_OVERRIDE_ATTRIB)
    # TODO add env override support env_kv = ma.get('Atom.Environ', '')
    if not task_arn:
        task_arn = TASKS_BY_JOB_SIZE.get(job_size)
    if not task_arn:
        publish_cloudwatch('job_size_mapping_failure', 1,
                           job_size=job_size, ecs_cluster_id=ECS_CLUSTER_ID)
        logging.error('Missing job_size mapping %s', job_size)
        # don't ack, in case a separate worker _does_ have the updated mapping
        # TODO re-publish with a fail marker incremented
        return

    script = m['Body']
    task_resp = ecs.run_task(
        cluster=ECS_CLUSTER_ID,
        taskDefinition=task_arn,
        overrides={
            'name': SCRIPTING_CONTAINER_NAME,
            'containerOverrides': [
                {
                    'command': [
                        'bash',
                        '-ec',
                        script,
                    ]
                }
            ]
        },
        count=1,
        startedBy=task_link,
        launchType='FARGATE',
    )
    task_fails = task_resp['failures']
    for task_fail in task_fails:
        publish_cloudwatch('job_run_task_fail', 1,
                           job_size=job_size, ecs_cluster_id=ECS_CLUSTER_ID)
        logging.error('Bogus task submission; failed_arm="%s" reason="%s"',
                      task_fail['arn'], task_fail['reason'])
    if task_fails:
        return  # don't ACK it, in hopes fail is temp

    logging.debug('ACK message_id="%s"', m_id)
    sqs.delete_message(QueueUrl=JOB_Q, ReceiptHandle=r_handle)


def main():
    while True:
        resp = sqs.receive_message(
            QueueUrl=JOB_Q,
            AttributeNames=[
                'All',
            ],
            MaxNumberOfMessages=1,
            MessageAttributeNames=[
                'All',
            ],
            VisibilityTimeout=5,
            WaitTimeSeconds=0,
        )
        for m in resp['Messages']:
            on_message(m)


if __name__ == '__main__':
    main()
